/* dds2_excite_test.c
 * The example program tests the analog outputs of the PBDDS-II.
 * This program will produce a 1 MHz signal on the oscilloscope that is on
 * for 10 microseconds, off for 10 microseconds on both channels, and then repeat this pattern.
 * IRQ0 is enabled.
 *
 * SpinCore Technologies, Inc.
 * www.spincore.com
 * $Date: 2008/06/13 13:36:00 $
 */

#include <stdio.h>
#include "spinapi.h"

#define CLOCK_FREQ 75.0

// User friendly names for the control bits
#define TX_ENABLE 1
#define TX_DISABLE 0
#define PHASE_RESET 1
#define NO_PHASE_RESET 0
#define DO_TRIGGER 1
#define NO_TRIGGER 0
#define FREQ0 0
#define FREQ1 1
#define AMP0 0
#define AMP1 1
#define AMP2 2
#define AMP3 3
#define AMP4 4
#define PHASE0 0

/**************************************************************************
* This example demonstrates the ability to change the "return-to" states  *
*      of the TTL Flag bits                                               *
***************************************************************************/
int main(int argc, char *argv[])
{
    printf ("Using spinapi library version %s\n", pb_get_version());
    pb_set_debug(1);
    if(pb_count_boards()<=0)
    {
        printf("No boards were detected in your system. Please check your connections.\n");
        system("pause");       
        return -1;
    }
    
    if(pb_init()) {
        printf ("Error initializing board: %s\n", pb_get_error());    
	    system("pause");
        return -1;
    }
    
    pb_set_clock(75.0); //Set the PB core clock value.
    pb_write_register(0x40000+0x08,0x000); 
    
/****** THIS SECTION PROGRAMS REGISTERS FOR CHANNEL 1 (DDS0) ******/    
    pb_select_dds(0); //Select DDS0 (this is selected by default.)
    
    pb_start_programming(FREQ_REGS); //Program DDS0's frequency registers. NOTE: Each Channel has 16 available Frequency registers
        pb_set_freq(1.0*MHz); //Program Frequency Register 0.
        pb_set_freq(2.0*MHz); //Program Frequency Register 1.
    pb_stop_programming();

    pb_start_programming(TX_PHASE_REGS); //Program DDS0's phase registers. NOTE: Each Channel has 8 available Phase registers
        pb_set_phase(0.0);
    pb_stop_programming();
 
    pb_set_amp(0.2,0); //Program DDS0's amplitude registers.  /** THIS IS WHAT YOU MUST CHANGE TO ADJUST THE AMPLITUDE OF CHANNEL 1 TO MATCH CHANNEL 2 **/
    pb_set_amp(0.5,1);
    pb_set_amp(0.7,2);
    pb_set_amp(1.0,3);
    // NOTE: Each Channel has 4 available amplitude registers
/**********************************************************************/ 
 
/***** THIS SECTION PROGRAMS REGISTERS FOR CHANNEL 2 (DDS1) *****/    
    pb_select_dds(1); //Select DDS1.

    pb_start_programming(FREQ_REGS); //Program DDS1's frequency registers. 
        pb_set_freq(1.0*MHz); //Program Frequency Register 0.
        pb_set_freq(2.0*MHz); //Program Frequency Register 1.
    pb_stop_programming();

    pb_start_programming(TX_PHASE_REGS); //Program DDS1's phase registers. 
        pb_set_phase(90.0);
    pb_stop_programming();

    pb_set_amp(0.2,0); //Program DDS1's amplitude registers.
    pb_set_amp(0.5,1);
    pb_set_amp(0.7,2);
    pb_set_amp(1.0,3);
/************************************************************************/

//The next section is where the actual pulse program is written:
//The line below shows which parameter/argument corresponds to which register or flag.
/***** SPINCORE_API int pb_inst_dds2(int freq0, int phase0, int amp0, int dds_en0, int phase_reset0, int freq1, int phase1, int amp1, int dds_en1, int phase_reset1, int flags, int inst, int inst_data, double length) ****/

 pb_start_programming(PULSE_PROGRAM);  //This line is used to start programming the pulse program.

/********************************************************************************************************
* Each piece of the pulse program is specified by a pb_inst_dds2(...) command                           *
* The first line says that for Channel 1 use:                                                           *
*        Frequency register 0 (which was programmed to be 1 MHz)                                        *
*        Phase register 0 (which was programmed to be 0)                                                *        
*        Amplitude register 0 (which I changed to be 0.85 in order to match the amplitude of channel 2) *
*        Enable Channel 1                                                                               *
*        Do not reset the phase of channel 1                                                            *
* The first line also says for Channel 2 use:                                                           *
*        Frequency register 0 (which was programmed to be 1 MHz)                                        *
*        Phase register 0 (which was programmed to be 0)                                                *
*        Amplitude register 0 (which was programmed to be 1.0)                                          *
*        Do not enable Channel                                                                          *
*        Reset the phase of channel 1                                                                   *
* 0xfff is a hexadecimal value that corresponds to the TTL flags fff equals 1111 1111 1111 which means  *
*       that all 12 bits are HIGH                                                                       *
* The specific instruction for this line is CONTINUE which does not require a DATA field so it is set   *
*       to zero.                                                                                        *
* All of this takes place for 10us                                                                      *
* The second line does the same thing except Channel 1 and Channel 2 are switched, the TTL flags are    *
*       now all LOW, and the instruction branches back to the first line (this also happens for 10us)   *
********************************************************************************************************/   

       int start = pb_inst_dds2(FREQ0, PHASE0, AMP3, TX_ENABLE,  NO_PHASE_RESET, FREQ0, PHASE0, AMP0, TX_DISABLE, PHASE_RESET, 0xfff, CONTINUE, 0, 10.0*us);
                   pb_inst_dds2(FREQ0, PHASE0, AMP0, TX_DISABLE, PHASE_RESET, FREQ0, PHASE0, AMP3, TX_ENABLE,  NO_PHASE_RESET, 0x000, BRANCH, start, 10.0*us);
                   
                            
/*********************************************************************************************************
* The section below specifies the interrupt sub routine                                                  *
* The instructions follow the same general structure except that LOOP, END_LOOP and RTI instructions     *
*     are used.  Here you will notice the error where Channel 1 uses Frequency register 1, but Channel 2 *
*     uses Frequency register 0, so they are not the same frequency                                      *
*********************************************************************************************************/               
       int isr1 = pb_inst_dds2(FREQ1, PHASE0, AMP0, TX_ENABLE, NO_PHASE_RESET, FREQ1, PHASE0, AMP0, TX_ENABLE, NO_PHASE_RESET, 0xfff, LOOP, 10, 200.0*us);
                  pb_inst_dds2(FREQ1, PHASE0, AMP0, TX_DISABLE, PHASE_RESET, FREQ0, PHASE0, AMP0, TX_DISABLE, PHASE_RESET, 0x000, END_LOOP, isr1, 200.0*us);
                  pb_inst_dds2(FREQ0, PHASE0, AMP0, TX_DISABLE, PHASE_RESET, FREQ0, PHASE0, AMP0, TX_DISABLE, PHASE_RESET, 0x000, RTI, isr1, 100.0*us);
 pb_stop_programming();  // This line ends the pulse program
 
 
// The line below shows how to change the return to state for the TTL flag bits once the PB Core has stopped running
// Starting location is located at address 0x40000 + 0x08
 #define FLAG_STATES (0x40000+0x08)
 pb_write_register(FLAG_STATES,0xA5A); // You may choose any 12 bit number to represent the TTL Flags
 
 
 
 /***  The section below sets the interrupt masks ***/
 pb_set_isr(0, isr1);               //Program IRQ0 ISR
 pb_set_irq_enable_mask(0x1);       //Interrupt enable mask. 0x1 enables IRQ0.
 pb_set_irq_immediate_mask(0x0);    //Set which IRQs are immediate IRQs (immediately perform ISR)
	
	printf("\nThis example shows the ability to change the 'return-to' states of the TTL \n ");
    printf("FLag bits with only one simple write to the board.\n\n");
	printf("IRQ0 will cause an ISR to be performed that produces a 2.0 MHz signal on both\n channels that is on for 200.0 us, off for 200.0 us and repeats this pattern\n 10 times.\nIt will turn both channels off for 100.0 us and then returns to the main\n pulse program.\n\n");
	
    pb_start();
    system("pause");
    pb_stop(); 
    
	return 0;
}

